/* XGContextWindows - DPS like methods for window/screen handling

   Copyright (C) 1999 Free Software Foundation, Inc.

   Written by:  Adam Fedor <fedor@gnu.org>
   Date: Nov 1999
   
   This file is part of GNUstep

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.
   
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
   
   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   */

#include <math.h>
#include <Foundation/NSString.h>
#include <Foundation/NSArray.h>
#include <Foundation/NSValue.h>
#include <AppKit/NSGraphics.h>
#include <AppKit/GSWraps.h>

#ifdef XDPS_BACKEND_LIBRARY
#define XGContext       NSDPSContext
#include "gnustep/xdps/NSDPSContextWindow.h"
#include "PXKDrawingEngine.h"
#else
#include "gnustep/xgps/XGContextWindow.h"
#include "gnustep/xgps/XGContextPrivate.h"
#include "XGDrawingEngine.h"
#endif
#include <X11/cursorfont.h>

#define	OVERRIDE	0

#define MAX_SCREENS	100
#define XDPY (((RContext *)context)->dpy)
#define XDRW (((RContext *)context)->drawable)
#define XSCR (((RContext *)context)->screen_number)

/* This works around problems with non-"identity" matrices */
#ifdef XDPS_BACKEND_LIBRARY
#define DPSNormalMatrix \
  do { \
    if ([[NSUserDefaults standardUserDefaults] boolForKey: @"DPSDefaultMatrix"] == NO) \
      { \
        float nctm[6]; \
        NSDebugLog(@"Resetting default matrix\n"); \
        nctm[0]=1; nctm[1]=0; nctm[2]=0; nctm[3]=-1; nctm[4]=0; nctm[5]=0; \
        GSWSetMatrix(self, nctm); \
      } \
    } while (0)
#else
#define DPSNormalMatrix
#endif

/*
 * Name for application root window.
 */
static char	*rootName = 0;

/*
 * Is this running under the WindowMaker window manager?
 */
static BOOL	windowMaker = NO;

/*
 * Keep track of the windows we control
 */
static NSMapTable* windowmaps = NULL;
static NSMapTable* windowtags = NULL;

/* Current mouse grab window */
static gswindow_device_t *grab_window = NULL;

/*
 * The next two functions derived from WindowMaker by Alfredo K. Kojima
 */
static unsigned char*
PropGetCheckProperty(Display *dpy, Window window, Atom hint, Atom type,
		     int format, int count, int *retCount)
{
  Atom type_ret;
  int fmt_ret;
  unsigned long nitems_ret;
  unsigned long bytes_after_ret;
  unsigned char *data;
  int tmp;

  if (count <= 0)
    tmp = 0xffffff;
  else
    tmp = count;

  if (XGetWindowProperty(dpy, window, hint, 0, tmp, False, type,
			 &type_ret, &fmt_ret, &nitems_ret, &bytes_after_ret,
			 (unsigned char **)&data)!=Success || !data)
    return NULL;

  if ((type!=AnyPropertyType && type!=type_ret)
    || (count > 0 && nitems_ret != count)
    || (format != 0 && format != fmt_ret))
    {
      XFree(data);
      return NULL;
    }

  if (retCount)
    *retCount = nitems_ret;

  return data;
}

static XGWMType
checkWindowManager(Display *dpy)
{
  Atom	*data;
  Atom	atom;
  Atom	noticeboard;
  int	count;

  atom = XInternAtom(dpy, "_WINDOWMAKER_WM_PROTOCOLS", False);
  noticeboard = XInternAtom(dpy, "_WINDOWMAKER_NOTICEBOARD", False);

  data = (Atom*)PropGetCheckProperty(dpy, DefaultRootWindow(dpy), atom,
    XA_ATOM, 32, -1, &count);

  if (data == 0)
    {
      windowMaker = NO;
    }
  else
    {
      int	i = 0;

      while (i < count && data[i] != noticeboard)
	{
	  i++;
	}
      XFree(data);

      if (i < count)
	{
	  Window	*win;
	  void		*d;

	  win = (Window*)PropGetCheckProperty(dpy, DefaultRootWindow(dpy), 
	    noticeboard, XA_WINDOW, 32, -1, &count);

	  if (win == 0)
	    {
	      windowMaker = NO;
	    }
	  else
	    {
	      d = PropGetCheckProperty(dpy, *win, noticeboard,
		XA_WINDOW, 32, 1, NULL);
	      if (d == 0)
		{
		  windowMaker = NO;
		}
	      else
		{
		  XFree(d);
		  windowMaker = YES;
		}
	    }
	}
      else
	{
	  windowMaker = YES;
	}
    }
  if (windowMaker)
    return XGWM_WINDOWMAKER;
  else
    return XGWM_UNKNOWN;
}

static void
setNormalHints(Display *d, gswindow_device_t *w)
{
  if (w->siz_hints.flags & (USPosition | PPosition))
    NSDebugLLog(@"Hints", @"Hint posn %d: %d, %d",
      w->number, w->siz_hints.x, w->siz_hints.y);
  if (w->siz_hints.flags & (USSize | PSize))
    NSDebugLLog(@"Hints", @"Hint size %d: %d, %d",
      w->number, w->siz_hints.width, w->siz_hints.height);
  if (w->siz_hints.flags & PMinSize)
    NSDebugLLog(@"Hints", @"Hint mins %d: %d, %d",
      w->number, w->siz_hints.min_width, w->siz_hints.min_height);
  if (w->siz_hints.flags & PMaxSize)
    NSDebugLLog(@"Hints", @"Hint maxs %d: %d, %d",
      w->number, w->siz_hints.max_width, w->siz_hints.max_height);
  if (w->siz_hints.flags & PResizeInc)
    NSDebugLLog(@"Hints", @"Hint incr %d: %d, %d",
      w->number, w->siz_hints.width_inc, w->siz_hints.height_inc);
  XSetNormalHints(d, w->ident, &w->siz_hints);
}

@interface NSEvent (WindowHack)
- (void) _patchLocation: (NSPoint)loc;
@end

@implementation NSEvent (WindowHack)
- (void) _patchLocation: (NSPoint)loc
{
  location_point = loc;
}
@end

@implementation XGContext (DPSWindow)

+ (gswindow_device_t *) _windowForXWindow: (Window)xWindow
{
  return NSMapGet(windowmaps, (void *)xWindow);
}

+ (gswindow_device_t *) _windowWithTag: (int)windowNumber
{
  return NSMapGet(windowtags, (void *)windowNumber);
}

/*
 * Convert a window frame in OpenStep absolute screen coordinates to
 * a frame in X absolute screen coordinates by flipping an applying
 * offsets to allow for the X window decorations.
 */
- (NSRect) _OSFrameToXFrame: (NSRect)o for: (void*)window
{
  gswindow_device_t	*win = (gswindow_device_t*)window;
  NSRect	x;
  float	t, b, l, r;

  [self DPSstyleoffsets: &l : &r : &t : &b : win->win_attrs.window_style];

  x.size.width = o.size.width - l - r;
  x.size.height = o.size.height - t - b;
  x.origin.x = o.origin.x + l;
  x.origin.y = o.origin.y + o.size.height - t;
  x.origin.y = DisplayHeight(XDPY, win->screen) - x.origin.y;
NSDebugLLog(@"Frame", @"O2X %d, %@, %@", win->number,
  NSStringFromRect(o), NSStringFromRect(x));
  return x;
}

/*
 * Convert a window frame in OpenStep absolute screen coordinates to
 * a frame suitable for setting X hints for a window manager.
 */
- (NSRect) _OSFrameToXHints: (NSRect)o for: (void*)window
{
  gswindow_device_t	*win = (gswindow_device_t*)window;
  NSRect	x;
  float	t, b, l, r;

  [self DPSstyleoffsets: &l : &r : &t : &b : win->win_attrs.window_style];

  x.size.width = o.size.width - l - r;
  x.size.height = o.size.height - t - b;
  x.origin.x = o.origin.x;
  x.origin.y = o.origin.y + o.size.height;
  x.origin.y = DisplayHeight(XDPY, win->screen) - x.origin.y;
NSDebugLLog(@"Frame", @"O2H %d, %@, %@", win->number,
  NSStringFromRect(o), NSStringFromRect(x));
  return x;
}

/*
 * Convert a window frame in X absolute screen coordinates to a frame
 * in OpenStep absolute screen coordinates by flipping an applying
 * offsets to allow for the X window decorations.
 */
- (NSRect) _XFrameToOSFrame: (NSRect)x for: (void*)window
{
  gswindow_device_t	*win = (gswindow_device_t*)window;
  NSRect	o;
  float	t, b, l, r;

  [self DPSstyleoffsets: &l : &r : &t : &b : win->win_attrs.window_style];
  o.size.width = x.size.width + l + r;
  o.size.height = x.size.height + t + b;
  o.origin.x = x.origin.x - l;
  o.origin.y = DisplayHeight(XDPY, win->screen) - x.origin.y;
  o.origin.y = o.origin.y - o.size.height + t;
NSDebugLLog(@"Frame", @"X2O %d, %@, %@", win->number,
  NSStringFromRect(x), NSStringFromRect(o));
  return o;
}

- (gswindow_device_t *)_rootWindowForScreen: (int)screen
{
  int x, y, width, height;
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)screen);
  if (window)
    return window;

  window = objc_malloc(sizeof(gswindow_device_t));

  window->screen = screen;
  window->ident  = RootWindow(XDPY, screen);
  window->root   = window->ident;
  window->type   = NSBackingStoreNonretained;
  window->number = screen;
  window->map_state = IsViewable;
  if (window->ident)
    XGetGeometry(XDPY, window->ident, &window->root, 
		 &x, &y, &width, &height,
		 &window->border, &window->depth);

  window->xframe = NSMakeRect(x, y, width, height);
  NSMapInsert (windowtags, (void*)window->number, window);
  NSMapInsert (windowmaps, (void*)window->ident,  window);
  return window;
}

/* Create the window and screen list if necessary, add the root window to
   the window list as window 0 */
- (void) _checkWindowlist
{
  if (windowmaps)
    return;

  windowmaps = NSCreateMapTable(NSNonOwnedPointerMapKeyCallBacks,
				 NSNonOwnedPointerMapValueCallBacks, 20);
  windowtags = NSCreateMapTable(NSIntMapKeyCallBacks,
				 NSNonOwnedPointerMapValueCallBacks, 20);
}

/* Sets up a backing pixmap when a window is created or resized.  This is
   only done if the Window is buffered or retained. */
- (void) _createBuffer: (gswindow_device_t *)window resize: (BOOL)resize
{
  if (window->type == NSBackingStoreNonretained)
    return;
  if (!resize && window->buffer)
    return;

  /* If this was a resize, then create a new pixmap and free the old one.
     I assume a resize event will trigger a redisplay of the data, and we
     don't need to keep any information from the old pixmap. */
  if (resize && window->buffer)
    XFreePixmap(XDPY, window->buffer);
  if (resize && window->alpha_buffer)
    {
      XFreePixmap(XDPY, window->alpha_buffer);
      window->alpha_buffer = 0;
    }

  if (window->depth == 0)
    window->depth = DefaultDepth(XDPY, XSCR);
  if (NSWidth(window->xframe) == 0 && NSHeight(window->xframe) == 0)
    {
      NSDebugLLog(@"NSWindow", @"Cannot create buffer for ZeroRect frame");
      window->buffer = 0;
      return;
    }

  window->buffer = XCreatePixmap(XDPY, window->root,
				 NSWidth(window->xframe),
				 NSHeight(window->xframe),
				 window->depth);

  if (!window->buffer)
    {
      NSLog(@"DPS Windows: Unable to create backing store\n");
      return;
    }

  XFillRectangle(XDPY,
		 window->buffer,
		 window->gc,
		 0, 0, 
		 NSWidth(window->xframe),
		 NSHeight(window->xframe));

}

- (void) _setupRootWindow
{
  NSProcessInfo		*pInfo = [NSProcessInfo processInfo];
  NSString		*n = [[NSProcessInfo processInfo] processName];
  NSArray		*args;
  unsigned		i;
  int			argc;
  char			**argv;
  XClassHint		classhint; 
  GNUstepWMAttributes	win_attrs;
  XTextProperty		windowName;

  /*
   * Initialize time of last events to be the start of time - not
   * the current time!
   */
  if (CurrentTime == 0)
    {
      generic.lastClick = 1;
      generic.lastMotion = 1;
      generic.lastTime = 1;
    }

  /*
   * Set up standard atoms.
   */
  generic.protocols_atom = XInternAtom(XDPY, "WM_PROTOCOLS", False);
  generic.take_focus_atom = XInternAtom(XDPY, "WM_TAKE_FOCUS", False);
  generic.delete_win_atom = XInternAtom(XDPY, "WM_DELETE_WINDOW", False);
  generic.miniaturize_atom
    = XInternAtom(XDPY, "_GNUSTEP_WM_MINIATURIZE_WINDOW", False);
  generic.win_decor_atom = XInternAtom(XDPY,"_GNUSTEP_WM_ATTR", False);
  generic.titlebar_state_atom
    = XInternAtom(XDPY, "_GNUSTEP_TITLEBAR_STATE", False);

  /*
   * determine window manager in use.
   */
  generic.wm = checkWindowManager(XDPY);

  /*
   * Mark this as a GNUstep app with the current application name.
   */
  if (rootName != 0)
    {
      objc_free((void*)rootName);
    }
  i = [n cStringLength];
  rootName = objc_malloc(i+1);
  [n getCString: rootName];
  classhint.res_name = rootName;
  classhint.res_class = "GNUstep";
  XSetClassHint(XDPY, XDRW, &classhint);

  /*
   * Set app name as root window title - probably unused unless
   * the window manager wants to keep us in a menu or something like that.
   */
  XStringListToTextProperty((char**)&classhint.res_name, 1, &windowName);
  XSetWMName(XDPY, XDRW, &windowName);
  XSetWMIconName(XDPY, XDRW, &windowName);

  /*
   * Record the information used to start this app.
   */
  args = [pInfo arguments];
  argc = [args count];
  argv = (char**)objc_malloc(argc*sizeof(char*));
  for (i = 0; i < argc; i++)
    {
      argv[i] = (char*)[[args objectAtIndex: i] cString];
    }
  XSetCommand(XDPY, XDRW, argv, argc);
  objc_free(argv);

  /*
   * Tell WindowMaker not to set up an app icon for us - we'll make our own.
   */
  win_attrs.flags |= GSExtraFlagsAttr;
  win_attrs.extra_flags |= GSNoApplicationIconFlag;
  if (generic.wm == XGWM_WINDOWMAKER)
    {
      XChangeProperty(XDPY, XDRW,
	generic.win_decor_atom, generic.win_decor_atom,
	32, PropModeReplace, (unsigned char *)&win_attrs,
	sizeof(GNUstepWMAttributes)/sizeof(CARD32));
    }
}

/* Fill in various window parameters. */
- (void) _updateWindowParameters: (gswindow_device_t *)window
{
  int      x, y;
  unsigned width, height;
  unsigned old_width = NSWidth(window->xframe);
  unsigned old_height = NSHeight(window->xframe);
  XWindowAttributes winattr;
  
  if (!window->ident)
    return;

  XFlush (XDPY);

  /* hack:
   * wait until a resize of window is finished (especially for NSMenu)
   * is there any way to wait until X finished it's stuff?
   * XSync(), XFlush() doesn't do the job!
   */
  { 
    int	i = 0;
    do
      {
	XGetGeometry(XDPY, window->ident, &window->root,
		     &x, &y, &width, &height,
		     &window->border, &window->depth);
      }
    while( i++<10 && height != window->siz_hints.height );
  }
  window->xframe.size.width = width;
  window->xframe.size.height = height;

  XGetWindowAttributes(XDPY, window->ident, &winattr);
  window->map_state = winattr.map_state;
  
  NSDebugLLog (@"NSWindow", @"window geom device ((%f, %f), (%f, %f))",
	        window->xframe.origin.x,  window->xframe.origin.y, 
	        window->xframe.size.width,  window->xframe.size.height);
  
  if (window->buffer && (old_width != width || old_height != height))
    {
      XFreePixmap(XDPY, window->buffer);
      window->buffer = 0;
      if (window->alpha_buffer)
	XFreePixmap(XDPY, window->alpha_buffer);
      window->alpha_buffer = 0;
    }
}

- (void) DPSwindow: (float)x : (float)y : (float)w : (float)h : (int)type : (int *)num
{
  static int		last_win_num = 0;
  int			num_protocols = 0;
  gswindow_device_t	*window;
  gswindow_device_t	*root;
  XGCValues		values;
  unsigned long		valuemask;
  XClassHint		classhint; 

  [self _checkWindowlist];
  root = [self _rootWindowForScreen: XSCR];

  /* We're not allowed to create a zero rect window */
  if (w == 0 && h == 0)
    {
      w = 200;
      h = 200;
    }
  /* Translate to X coordinates */
  y = DisplayHeight(XDPY, XSCR) - (y + h);

  window = objc_malloc(sizeof(gswindow_device_t));
  memset(window, '\0', sizeof(gswindow_device_t));
  window->screen = XSCR;
  window->xframe = NSMakeRect(x, y, w, h);
  window->type = type;
  window->root = root->ident;
  window->parent = root->ident;
  window->depth = ((RContext *)context)->depth;
  window->xwn_attrs.border_pixel = ((RContext *)context)->black;
  window->xwn_attrs.background_pixel = ((RContext *)context)->white;
  window->xwn_attrs.colormap = ((RContext *)context)->cmap;

  window->ident = XCreateWindow(XDPY, window->root,
				x, y, w, h,
				0, 
				((RContext *)context)->depth,
				CopyFromParent,
				((RContext *)context)->visual,
				(CWColormap | CWBackPixel|CWBorderPixel),
				&window->xwn_attrs);

  /*
   * Mark this as a GNUstep app with the current application name.
   */
  classhint.res_name = rootName;
  classhint.res_class = "GNUstep";
  XSetClassHint(XDPY, window->ident, &classhint);

  window->xwn_attrs.save_under = False;
  window->xwn_attrs.override_redirect = False;
  window->map_state = IsUnmapped;


  // Create an X GC for the content view set it's colors
  values.foreground = window->xwn_attrs.background_pixel;
  values.background = window->xwn_attrs.background_pixel;
  values.function = GXcopy;
  valuemask = (GCForeground | GCBackground | GCFunction);
  window->gc = XCreateGC(XDPY, window->ident, valuemask, &values);

  // Set the X event mask
  XSelectInput(XDPY, window->ident, ExposureMask | KeyPressMask |
				KeyReleaseMask | ButtonPressMask |
			     ButtonReleaseMask | ButtonMotionMask |
			   StructureNotifyMask | PointerMotionMask |
			       EnterWindowMask | LeaveWindowMask |
			       FocusChangeMask | PropertyChangeMask |
			    ColormapChangeMask | KeymapStateMask |
			    VisibilityChangeMask);

  window->gen_hints.window_group = XDRW;
  window->gen_hints.initial_state = NormalState;
  window->gen_hints.input = False;
  window->gen_hints.flags = StateHint | InputHint | WindowGroupHint;
  XSetWMHints(XDPY, window->ident, &window->gen_hints);

  /*
   * Prepare size/position hints, but don't set them now - ordering
   * the window in should automatically do it.
   */
  window->siz_hints.x = x;
  window->siz_hints.y = y;
  window->siz_hints.width = w;
  window->siz_hints.height = h;
  window->siz_hints.flags = USPosition|PPosition|USSize|PSize;

  window->protocols[num_protocols++] = generic.take_focus_atom;
  window->protocols[num_protocols++] = generic.delete_win_atom;
  if (generic.wm == XGWM_WINDOWMAKER)
    {
      window->protocols[num_protocols++] = generic.miniaturize_atom;
    }
  XSetWMProtocols(XDPY, window->ident, window->protocols, num_protocols);

  window->exposedRects = [NSMutableArray new];
  window->region = XCreateRegion();
  window->buffer = 0;
  window->alpha_buffer = 0;

  /*
   * FIXME - should this be protected by a lock for thread safety?
   * generate a unique tag for this new window.
   */
  do
    {
      last_win_num++;
    }
  while (last_win_num == 0 || NSMapGet(windowtags, (void*)last_win_num) != 0);
  window->number = last_win_num;

  if (num)
    *num = window->number;

  // Insert window into the mapping
  NSMapInsert(windowmaps, (void*)window->ident, window);
  NSMapInsert(windowtags, (void*)window->number, window);
}

- (void) DPStermwindow: (int)num
{
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)num);
  if (!window)
    return;

  if (window->ident)
    {
      XDestroyWindow(XDPY, window->ident);
      XFreeGC (XDPY, window->gc);
      NSMapRemove(windowmaps, (void*)window->ident);
    }

  if (window->buffer)
    XFreePixmap (XDPY, window->buffer);
  if (window->alpha_buffer)
    XFreePixmap (XDPY, window->alpha_buffer);
  XDestroyRegion (window->region);
  RELEASE(window->exposedRects);
  NSMapRemove(windowtags, (void*)num);
  objc_free(window);
}

/*
 * Return the offsets between the window content-view and it's frame
 * depending on the window style.
 */
- (void) DPSstyleoffsets: (float *) l : (float *) r : (float *) t : (float *) b : (int) style ;
{
  /*
   * This should get window-manager specific decoration information somehow
   */
  switch (generic.wm)
    {
      case XGWM_WINDOWMAKER:
	if (style == NSBorderlessWindowMask
	  || (style & NSIconWindowMask) || (style & NSMiniWindowMask))
	  {
	    *l = *r = *t = *b = 0.0;
	  }
	else
	  {
	    *l = *r = *t = *b = 1;
	    if (NSResizableWindowMask & style)
	      {
		*b = 9;
	      }
	    if ((style & NSTitledWindowMask) || (style & NSClosableWindowMask)
	      || (style & NSMiniaturizableWindowMask))
	      {
		*t = 23;
	      }
	  }
	break;

      case XGWM_UNKNOWN:
      default:
	/*
	 * FIXME
	 * This should make a good guess - at the moment use no offsets.
	 */
	*l = *r = *t = *b = 0.0;
	break;
    }
}

- (void) DPSstylewindow: (int)style : (int) num
{
  gswindow_device_t	*window;

  window = NSMapGet(windowtags, (void *)num);
  if (!window)
    return;

  if (window->win_attrs.window_style != style
    || (window->win_attrs.flags & GSWindowStyleAttr) == 0)
    {
      if (OVERRIDE && window->win_attrs.window_level >= NSDockWindowLevel
	&& style == NSBorderlessWindowMask)
	{
	  if (window->xwn_attrs.save_under == False
	    || window->xwn_attrs.override_redirect == False)
	    {
	      window->xwn_attrs.save_under = True;
	      window->xwn_attrs.override_redirect = True;
	      XChangeWindowAttributes(XDPY, window->ident,
		(CWSaveUnder|CWOverrideRedirect), &window->xwn_attrs);
	    }
	}
      else
	{
	  if (window->xwn_attrs.save_under == True
	    || window->xwn_attrs.override_redirect == True)
	    {
	      window->xwn_attrs.save_under = False;
	      window->xwn_attrs.override_redirect = False;
	      XChangeWindowAttributes(XDPY, window->ident,
		(CWSaveUnder|CWOverrideRedirect), &window->xwn_attrs);
	    }
	}

      window->win_attrs.flags |= GSWindowStyleAttr;
      window->win_attrs.window_style = style;

      // send WindowMaker WM window style hints
      if (generic.wm == XGWM_WINDOWMAKER)
	{
	  XChangeProperty(XDPY, window->ident,
	    generic.win_decor_atom, generic.win_decor_atom,
	    32, PropModeReplace, (unsigned char *)&window->win_attrs,
	    sizeof(GNUstepWMAttributes)/sizeof(CARD32));
	}
    }
}

- (void) DPStitlewindow: (const char *)window_title : (int) num
{
  XTextProperty windowName;
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)num);
  if (!window)
    return;

  if (window_title && window->ident)
    {
      XStringListToTextProperty((char**)&window_title, 1, &windowName);
      XSetWMName(XDPY, window->ident, &windowName);
      XSetWMIconName(XDPY, window->ident, &windowName);
    }
}

- (void) DPSdocedited: (int)edited : (int) num
{
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)num);
  if (!window)
    return;

  window->win_attrs.flags |= GSExtraFlagsAttr;
  if (edited)
    {
      window->win_attrs.extra_flags |= GSDocumentEditedFlag;
    }
  else
    {
      window->win_attrs.extra_flags &= ~GSDocumentEditedFlag;
    }
  // send WindowMaker WM window style hints
  if (generic.wm == XGWM_WINDOWMAKER)
    {
      XChangeProperty(XDPY, window->ident,
	generic.win_decor_atom, generic.win_decor_atom,
	32, PropModeReplace, (unsigned char *)&window->win_attrs,
	sizeof(GNUstepWMAttributes)/sizeof(CARD32));
    }
}

- (void) DPSminiwindow: (int) num
{
  gswindow_device_t	*window;

  window = NSMapGet(windowtags, (void *)num);
  if (window == 0)
    {
      return;
    }
  // AT PRESENT - things seem to go wrong if we set an icon winiwindow -
  // I think that the WindowManager prevents us getting most of the events
  // in the miniwindow - so it doesn't know when it have been unmapped
  // and we end up trying to perform draw operations on an unmapped window!
#if 1
  /*
   * If we haven't already done so - set the icon window hint for this
   * window so that the GNUstep mniwindow is displayed (if supported).
   */
  if ((window->gen_hints.flags & IconWindowHint) == 0)
    {
      NSWindow		*nswin;

      nswin = GSWindowWithNumber(window->number);
      if (nswin != nil)
	{
	  int			iNum = [[nswin counterpart] windowNumber];
	  gswindow_device_t	*iconw = NSMapGet(windowtags, (void *)iNum);

	  if (iconw != 0)
	    {
	      window->gen_hints.flags |= IconWindowHint;
	      window->gen_hints.icon_window = iconw->ident;
	      XSetWMHints(XDPY, window->ident, &window->gen_hints);
	    }
	}
    }
#endif
  XIconifyWindow(XDPY, window->ident, XSCR);
}

- (gswindow_device_t *) _windowdevice: (int)num
{
  gswindow_device_t *window;

  [self _checkWindowlist];
  window = NSMapGet(windowtags, (void *)num);
  if (!window)
    [NSException raise: @"DPSinvalidparam" 
		 format: @"Invalid window number %d", num];

  [self _updateWindowParameters: window];
  if (window->buffer == 0)
    [self _createBuffer: window resize: NO];

  GSWinitcontext (self, num, window->gc, 
  		  (window->buffer) ? window->buffer : window->ident, 
		  0, NSHeight(window->xframe));
  DPSNormalMatrix;

  return window;
}

- (void) DPSwindowdevice: (int)num
{
  [self _windowdevice: num];
}

- (void) DPSwindowdeviceround: (int)num
{
  [self _windowdevice: num];
}

- (void) DPScurrentwindow: (int *)num
{
}

- (void) DPSorderwindow: (int)op : (int)otherWin : (int)winNum
{
  gswindow_device_t	*window;
  gswindow_device_t	*other;
  int		level;

  window = [XGContext _windowWithTag: winNum];
  if (window == NULL)
    {
      return;
    }
  level = window->win_attrs.window_level;
  if (otherWin != 0)
    {
      other = [XGContext _windowWithTag: otherWin];
      level = other->win_attrs.window_level;
    }
  else
    {
      other = NULL;
    }
  DPSsetwindowlevel(self, level, winNum);

  /*
   * When we are ordering a window in, we must ensure that the position
   * and size hints are set for the window - the window could have been
   * moved or resized by the window manager before it was ordered out,
   * in which case, we will have been notified of the new position, but
   * will not yet have updated the window hints, so if the window manager
   * looks at the existing hints when re-mapping the window it will
   * place the window in an old location.
   */
  if (op != NSWindowOut && window->map_state != IsViewable)
    {
      XMoveWindow(XDPY, window->ident, window->siz_hints.x,
	window->siz_hints.y);
      setNormalHints(XDPY, window);
    }

  switch (op)
    {
      case NSWindowBelow:
        if (other != 0)
	  {
	    XWindowChanges chg;
	    chg.sibling = other->ident;
	    chg.stack_mode = Below;
	    XReconfigureWMWindow(XDPY, window->ident, window->screen,
	      CWSibling|CWStackMode, &chg);
	  }
	else
	  {
	    XWindowChanges chg;
	    chg.stack_mode = Below;
	    XReconfigureWMWindow(XDPY, window->ident, window->screen,
	      CWStackMode, &chg);
	  }
	XMapWindow(XDPY, window->ident);
	break;

      case NSWindowAbove:
        if (other != 0)
	  {
	    XWindowChanges chg;
	    chg.sibling = other->ident;
	    chg.stack_mode = Above;
	    XReconfigureWMWindow(XDPY, window->ident, window->screen,
	      CWSibling|CWStackMode, &chg);
	  }
	else
	  {
	    XWindowChanges chg;
	    chg.stack_mode = Above;
	    XReconfigureWMWindow(XDPY, window->ident, window->screen,
	      CWStackMode, &chg);
	  }
	XMapWindow(XDPY, window->ident);
	break;

      case NSWindowOut:
	XUnmapWindow(XDPY, window->ident);
	break;
    }
  /*
   * When we are ordering a window in, we must ensure that the position
   * and size hints are set for the window - the window could have been
   * moved or resized by the window manager before it was ordered out,
   * in which case, we will have been notified of the new position, but
   * will not yet have updated the window hints, so if the window manager
   * looks at the existing hints when re-mapping the window it will
   * place the window in an old location.
   */
  if (op != NSWindowOut && window->map_state != IsViewable)
    {
      XMoveWindow(XDPY, window->ident, window->siz_hints.x,
	window->siz_hints.y);
      setNormalHints(XDPY, window);
    }

  XFlush(XDPY);
}

- (void) DPSmovewindow: (float)x : (float)y : (int)num
{
}

/* Sync window information with possible change in X window */
- (void) DPSupdatewindow: (int)win
{
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)win);
  if (!window)
    return;

  [self _updateWindowParameters: window];
  if (window->buffer == 0)
    {
      [self _createBuffer: window resize: YES];
    }

  GSWinitcontext (self, win, window->gc, 
  		  (window->buffer) ? window->buffer : window->ident, 
		  0, NSHeight(window->xframe));
  DPSNormalMatrix;
}

- (void) DPSplacewindow: (float)x : (float)y : (float)w : (float)h : (int)win
{
  NSAutoreleasePool	*arp;
  NSRect		rect = NSMakeRect(x, y, w, h);
  NSRect		xVal;
  NSRect		last;
  NSEvent		*event;
  NSDate		*limit;
  NSMutableArray	*tmpQueue;
  unsigned		pos;
  float			xdiff;
  float			ydiff;
  gswindow_device_t	*window;
  NSWindow              *nswin;
  NSRect		frame;
  BOOL			resize = NO;
  BOOL			move = NO;

  window = [XGContext _windowWithTag: win];
  if (window == 0)
    return;
  nswin  = GSWindowWithNumber(win);
  frame = [nswin frame];
  if (NSEqualRects(rect, frame) == YES)
    return;
  if (NSEqualSizes(rect.size, frame.size) == NO)
    {
      resize = YES;
      move = YES;
    }
  if (NSEqualPoints(rect.origin, frame.origin) == NO)
    {
      move = YES;
    }
  xdiff = rect.origin.x - frame.origin.x;
  ydiff = rect.origin.y - frame.origin.y;

  xVal = [self _OSFrameToXHints: rect for: window];
  window->siz_hints.width = (int)xVal.size.width;
  window->siz_hints.height = (int)xVal.size.height;
  window->siz_hints.x = (int)xVal.origin.x;
  window->siz_hints.y = (int)xVal.origin.y;
  xVal = [self _OSFrameToXFrame: rect for: window];

  last = window->xframe;

  NSDebugLLog(@"Moving", @"Place %d - o:%@, x:%@", window->number,
    NSStringFromRect(rect), NSStringFromRect(xVal));
  
  XMoveResizeWindow (XDPY, window->ident,
    window->siz_hints.x, window->siz_hints.y,
    window->siz_hints.width, window->siz_hints.height);
  setNormalHints(XDPY, window);

  /*
   * Now massage all the events currently in the queue to make sure
   * mouse locations in our window are adjusted as necessary.
   */
  arp = [NSAutoreleasePool new];
  limit = [NSDate distantPast];	/* Don't wait for new events.	*/
  tmpQueue = [NSMutableArray arrayWithCapacity: 8];
  for (;;)
    {
      NSEventType	type;

      event = DPSGetEvent(self, NSAnyEventMask, limit,
	NSEventTrackingRunLoopMode);
      if (event == nil)
	break;
      type = [event type];
      if (type == NSAppKitDefined && [event windowNumber] == win)
	{
	  GSAppKitSubtype	sub = [event subtype];

	  /*
	   * Window movement or resize events for the window we are
	   * watching are posted immediately, so they can take effect
	   * before the placewindow returns.
	   */
	  if (sub == GSAppKitWindowMoved || sub == GSAppKitWindowResized)
	    {
	      [nswin sendEvent: event];
              if (sub == GSAppKitWindowResized)
                [nswin update];
	    }
	  else
	    {
	      [tmpQueue addObject: event];
	    }
	}
      else if (type != NSPeriodic && type != NSLeftMouseDragged
	&& type != NSRightMouseDragged && type != NSMouseMoved)
	{
	  /*
	   * Save any events that arrive before our window is moved - excepting
	   * periodic events (which we can assume will be outdated) and mouse
	   * movement events (which might flood us).
	   */
	  [tmpQueue addObject: event];
	}
      if (NSEqualRects(xVal, window->xframe) == YES ||
	NSEqualRects(rect, [nswin frame]) == YES)
	{
	  break;
	}
      if (NSEqualRects(last, window->xframe) == NO)
	{
	  NSDebugLog(@"From: %@\nWant %@\nGot %@",
	    NSStringFromRect(last),
	    NSStringFromRect(xVal),
	    NSStringFromRect(window->xframe));
	  last = window->xframe;
	}
    }
  /*
   * If we got any events while waiting for the window movement, we
   * may need to adjust their locations to match the new window position.
   */
  pos = [tmpQueue count];
  while (pos-- > 0) 
    {
      event = [tmpQueue objectAtIndex: pos];
      if ([event windowNumber] == win)
	{
	  NSPoint	loc = [event locationInWindow];

	  loc.x -= xdiff;
	  loc.y -= ydiff;
	  [event _patchLocation: loc];
	}
      DPSPostEvent(self, event, YES);
    }
  [arp release];
  /*
   * Failsafe - if X hasn't told us it has moved/resized the window, we
   * fake the notification and post them immediately, so they can take
   * effect before the placewindow returns.
   */
  if (NSEqualRects([nswin frame], rect) == NO)
    {
      NSEvent	*e;

      if (resize == YES)
	{
	  NSDebugLLog(@"Moving", @"Fake size %d - %@", window->number,
	    NSStringFromSize(rect.size));
	  e = [NSEvent otherEventWithType: NSAppKitDefined
				 location: NSZeroPoint
			    modifierFlags: 0
				timestamp: 0
			     windowNumber: win
				  context: self
				  subtype: GSAppKitWindowResized
				    data1: rect.size.width
				    data2: rect.size.height];
	  [nswin sendEvent: e];
	  [nswin update];
	}
      if (move == YES)
	{
	  NSDebugLLog(@"Moving", @"Fake move %d - %@", window->number,
	    NSStringFromPoint(rect.origin));
	  e = [NSEvent otherEventWithType: NSAppKitDefined
				 location: NSZeroPoint
			    modifierFlags: 0
				timestamp: 0
			     windowNumber: win
				  context: self
				  subtype: GSAppKitWindowMoved
				    data1: rect.origin.x
				    data2: rect.origin.y];
	  [nswin sendEvent: e];
	}
    }
}

- (void) DPSfrontwindow: (int *)num
{
}

- (void) DPSfindwindow: (float)x : (float)y : (int)op : (int)otherWin : (float *)lx : (float *)ly : (int *)winFound : (int *)didFind
{
}

- (void) DPScurrentwindowbounds: (int)num : (float *)x : (float *)y : (float *)w : (float *)h
{
  gswindow_device_t *window;
  int screenHeight;

  [self _checkWindowlist];
  window = NSMapGet(windowtags, (void *)num);
  if (!window && num < MAX_SCREENS)
    window = [self _rootWindowForScreen: num];
  if (!window)
    return;

  if (window->buffer == 0 && window->map_state != IsViewable)
    {
      /* Sort of a hack. Prevent routines from thinking they can
         get pixel information from a window that is unmapped */
     if (x) *x = 0;
     if (y) *y = 0;
     if (w) *w = 0;
     if (h) *h = 0;
     return;
    }

  screenHeight = DisplayHeight(XDPY, window->screen);
  if (x)
    *x = NSMinX(window->xframe);
  if (y)
    *y = screenHeight - NSMaxY(window->xframe);
  if (w)
    *w = NSWidth(window->xframe);
  if (h)
    *h = NSHeight(window->xframe);
}

- (void) DPSsetexposurecolor
{
}

- (void) DPSsetsendexposed: (int)truth : (int)num
{
}

- (void) DPSsetautofill: (int)truth : (int)num
{
}

- (void) DPScurrentwindowalpha: (int)win : (int *)alpha
{
}

- (void) DPScountscreenlist: (int)ctxt : (int *)count
{
  int c = ScreenCount(XDPY);
  if (count)
    *count = c;
}

- (void) DPSscreenlist: (int)ctxt : (int)count : (int *)windows
{
  /* I guess screen numbers are in order starting from zero, but we
     put the main screen first */
 int i, j;
 if (count > 0)
   windows[0] = XSCR;
 j = 1;
 for (i = 0; i < count; i++)
   {
     if (i != XSCR)
       windows[j++] = i;
   }
}

- (void) DPSsetowner: (int)owner : (int)win
{
}

- (void) DPScurrentowner: (int)win : (int *)owner
{
}

- (void) DPSsetwindowtype: (int)type : (int)win
{
}

- (void) DPSsetwindowlevel: (int)level : (int)win
{
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)win);
  if (!window)
    return;

  if (window->win_attrs.window_level != level
    || (window->win_attrs.flags & GSWindowLevelAttr) == 0)
    {
      if (OVERRIDE && level >= NSDockWindowLevel
	&& window->win_attrs.window_style == NSBorderlessWindowMask)
	{
	  if (window->xwn_attrs.save_under == False
	    || window->xwn_attrs.override_redirect == False)
	    {
	      window->xwn_attrs.save_under = True;
	      window->xwn_attrs.override_redirect = True;
	      XChangeWindowAttributes(XDPY, window->ident,
		(CWSaveUnder|CWOverrideRedirect), &window->xwn_attrs);
	    }
	}
      else
	{
	  if (window->xwn_attrs.save_under == True
	    || window->xwn_attrs.override_redirect == True)
	    {
	      window->xwn_attrs.save_under = False;
	      window->xwn_attrs.override_redirect = False;
	      XChangeWindowAttributes(XDPY, window->ident,
		(CWSaveUnder|CWOverrideRedirect), &window->xwn_attrs);
	    }
	}

      window->win_attrs.flags |= GSWindowLevelAttr;
      window->win_attrs.window_level = level;

      // send WindowMaker WM window style hints
      if (generic.wm == XGWM_WINDOWMAKER)
	{
	  XEvent	event;

	  /*
	   * First change the window properties so that, if the window
	   * is not mapped, we have stored the required info for when
	   * the WM maps it.
	   */
	  XChangeProperty(XDPY, window->ident,
	    generic.win_decor_atom, generic.win_decor_atom,
	    32, PropModeReplace, (unsigned char *)&window->win_attrs,
	    sizeof(GNUstepWMAttributes)/sizeof(CARD32));
	  /*
	   * Now send a message for rapid handling.
	   */
	  event.xclient.type = ClientMessage;
	  event.xclient.message_type = generic.win_decor_atom;
	  event.xclient.format = 32;
	  event.xclient.display = XDPY;
	  event.xclient.window = window->ident;
	  event.xclient.data.l[0] = GSWindowLevelAttr;
	  event.xclient.data.l[1] = window->win_attrs.window_level;
	  event.xclient.data.l[2] = 0;
	  event.xclient.data.l[3] = 0;
	  XSendEvent(XDPY, DefaultRootWindow(XDPY), False,
	    SubstructureRedirectMask, &event);
	}
    }
}

- (void) DPScurrentwindowlevel: (int)win : (int *)level
{
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)win);
  /*
   * If we have previously set a level for this window - return the value set.
   */
  if (window != 0 && (window->win_attrs.flags & GSWindowLevelAttr))
    *level = window->win_attrs.window_level;
}

- (void) DPScountwindowlist: (int)context : (int *)count
{
}

- (void) DPSwindowlist: (int)context : (int)count : (int *)windows
{
}

- (void) DPSsetwindowdepthlimit: (int)limit : (int)win
{
}

- (void) DPScurrentwindowdepthlimit: (int)win : (int *)limit
{
}

- (void) DPScurrentwindowdepth: (int)win : (int *)depth
{
  gswindow_device_t *window;

  [self _checkWindowlist];
  window = NSMapGet(windowtags, (void *)win);
  if (!window && win < MAX_SCREENS)
    window = [self _rootWindowForScreen: win];
  if (!window)
    return;

  if (depth)
    *depth = window->depth;
}

- (void) DPSsetdefaultdepthlimit: (int)limit
{
}

- (void) DPScurrentdefaultdepthlimit: (int *)limit
{
}

- (void) DPSsetmaxsize: (float)width : (float)height : (int)win
{
  gswindow_device_t	*window;
  NSRect		r;

  window = NSMapGet(windowtags, (void *)win);
  if (window == 0)
    {
      return;
    }
  r = NSMakeRect(0, 0, width, height);
  r = [self _OSFrameToXFrame: r for: window];
  window->siz_hints.flags |= PMaxSize;
  window->siz_hints.max_width = r.size.width;
  window->siz_hints.max_height = r.size.height;
  setNormalHints(XDPY, window);
}

- (void) DPSsetminsize: (float)width : (float)height : (int)win
{
  gswindow_device_t	*window;
  NSRect		r;

  window = NSMapGet(windowtags, (void *)win);
  if (window == 0)
    {
      return;
    }
  r = NSMakeRect(0, 0, width, height);
  r = [self _OSFrameToXFrame: r for: window];
  window->siz_hints.flags |= PMinSize;
  window->siz_hints.min_width = r.size.width;
  window->siz_hints.min_height = r.size.height;
  setNormalHints(XDPY, window);
}

- (void) DPSsetresizeincrements: (float)width : (float)height : (int)win
{
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)win);
  if (window == 0)
    {
      return;
    }
  window->siz_hints.flags |= PResizeInc;
  window->siz_hints.width_inc = width;
  window->siz_hints.height_inc = height;
  setNormalHints(XDPY, window);
}

// process expose event
- (void) _addExposedRectangle: (XRectangle)rectangle : (int)win
{
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)win);
  if (!window)
    return;

  if (window->type != NSBackingStoreNonretained)
    {
      XGCValues values;
      unsigned long valuemask;

      // window has a backing store so just copy the exposed rect from the
      // pixmap to the X window

      NSDebugLLog (@"NSWindow", @"copy exposed area ((%d, %d), (%d, %d))",
		  rectangle.x, rectangle.y, rectangle.width, rectangle.height);

      values.function = GXcopy;
      values.plane_mask = AllPlanes;
      values.clip_mask = None;
      values.foreground = ((RContext *)context)->white;
      valuemask = (GCFunction | GCPlaneMask | GCClipMask | GCForeground);
      XChangeGC(XDPY, window->gc, valuemask, &values);

      XCopyArea (XDPY, window->buffer, window->ident, window->gc,
		 rectangle.x, rectangle.y, rectangle.width, rectangle.height,
		 rectangle.x, rectangle.y);
    }
  else
    {
      NSRect	rect;

      // no backing store, so keep a list of exposed rects to be
      // processed in the _processExposedRectangles method
      // Add the rectangle to the region used in -_processExposedRectangles
      // to set the clipping path.
      XUnionRectWithRegion (&rectangle, window->region, window->region);

      // Transform the rectangle's coordinates to PS coordinates and add
      // this new rectangle to the list of exposed rectangles.
      rect.origin = NSMakePoint((float)rectangle.x, rectangle.y);
      rect.size = NSMakeSize(rectangle.width, rectangle.height);
      [window->exposedRects addObject: [NSValue valueWithRect: rect]];
    }
}

- (void) DPSflushwindowrect: (float)x : (float)y : (float)w : (float)h : (int)win
{
  int xi, yi, width, height;
  XGCValues values;
  unsigned long valuemask;
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)win);
  if (!window)
    return;

  if (window->type == NSBackingStoreNonretained)
    {
      XFlush(XDPY);
      return;
    }

  /* FIXME: Doesn't take into account any offset added to the window
     (from PSsetXgcdrawable) or possible scaling (unlikely in X-windows,
     but what about other devices?) */
  y = NSHeight(window->xframe) - (y+h);

  values.function = GXcopy;
  values.plane_mask = AllPlanes;
  values.clip_mask = None;
  valuemask = (GCFunction | GCPlaneMask | GCClipMask);
  XChangeGC(XDPY, window->gc, valuemask, &values);

  xi = x;		// width/height seems
  yi = y;		// to require +1 pixel
  width = w + 1;	// to copy out
  height = h + 1;

  NSDebugLLog (@"NSWindow", 
	       @"copy X rect ((%d, %d), (%d, %d))", xi, yi, width, height);

  if (width > 0 || height > 0)
    XCopyArea (XDPY, window->buffer, window->ident, window->gc, 
	       xi, yi, width, height, xi, yi);

  XFlush(XDPY);
}

// handle X expose events
- (void) _processExposedRectangles: (int)win
{
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)win);
  if (!window)
    return;

  if (window->type != NSBackingStoreNonretained)
    return;

  // Set the clipping path to the exposed rectangles
  // so that further drawing will not affect the non-exposed region
  XSetRegion (XDPY, window->gc, window->region);

  // We should determine the views that need to be redisplayed. Until we
  // fully support scalation and rotation of views redisplay everything.
  // FIXME: It seems wierd to trigger a front-end method from here...
  [GSWindowWithNumber(win) display];

  // Restore the exposed rectangles and the region
  [window->exposedRects removeAllObjects];
  XDestroyRegion (window->region);
  window->region = XCreateRegion();
  XSetClipMask (XDPY, window->gc, None);
}

- (void) DPScapturemouse: (int)win
{
  int ret;
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)win);
  if (!window)
    return;

  ret = XGrabPointer(XDPY, window->ident, False,
		     PointerMotionMask | ButtonReleaseMask | ButtonPressMask,
		     GrabModeAsync, GrabModeAsync, None, None, CurrentTime);

  if (ret != GrabSuccess)
    NSLog(@"Failed to grab pointer\n");
  else
    {
      grab_window = window;
      NSDebugLLog(@"NSWindow", @"Grabbed pointer\n");
    }
}

- (void) DPSreleasemouse
{
  XUngrabPointer(XDPY, CurrentTime);
  grab_window = NULL;
}

- (void) DPSsetinputfocus: (int)win
{
  gswindow_device_t *window = NSMapGet(windowtags, (void*)win);

  if (window == 0)
    {
      NSDebugLLog(@"Focus", @"Setting focus to unknown win %d", win);
      return;
    }
  /*
   * If we have an outstanding request to set focus to this window,
   * we don't want to do it again.
   */
  if (win == generic.desiredFocusWindow && generic.focusRequestNumber != 0)
    {
      NSDebugLLog(@"Focus", @"Resetting focus to %d", window->number);
    }
  else
    {
      NSDebugLLog(@"Focus", @"Setting focus to %d", window->number);
    }
  generic.desiredFocusWindow = win;
  generic.focusRequestNumber = XNextRequest(XDPY);
  XSetInputFocus(XDPY, window->ident, RevertToPointerRoot, generic.lastTime);
}

/*
 * Instruct window manager that the specified window is 'key', 'main', or
 * just a normal window.
 */
- (void) DPSsetinputstate: (int)win : (int)st
{
  if (generic.wm == XGWM_WINDOWMAKER)
    {
      gswindow_device_t *window = NSMapGet(windowtags, (void*)win);
      XEvent		event;

      event.xclient.type = ClientMessage;
      event.xclient.message_type = generic.titlebar_state_atom;
      event.xclient.format = 32;
      event.xclient.display = XDPY;
      event.xclient.window = window->ident;
      event.xclient.data.l[0] = st;
      event.xclient.data.l[1] = 0;
      event.xclient.data.l[2] = 0;
      event.xclient.data.l[3] = 0;
      XSendEvent(XDPY, DefaultRootWindow(XDPY), False,
	SubstructureRedirectMask, &event);
    }
}

- (void) DPScurrentserverdevice: (void **)serverptr
{
  static Display *ptrloc;
  ptrloc = XDPY;
  if (serverptr != NULL)
    *serverptr = &ptrloc;
}

- (void) DPScurrentwindowdevice: (int)win : (void **)windowptr
{
  static Window ptrloc;
  gswindow_device_t *window;

  window = NSMapGet(windowtags, (void *)win);
  if (window != NULL)
    ptrloc = window->ident;
  else
    ptrloc = 0;
  if (windowptr != NULL)
    *windowptr = &ptrloc;
}

/* Cursor Ops */
typedef struct _xgps_cursor_id_t {
  Cursor c;
  BOOL   defined;
} xgps_cursor_id_t;

static char xgps_blank_cursor_bits [] = {
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

static Cursor xgps_blank_cursor = None;
static BOOL   cursor_hidden = NO;

static void
xgps_set_cursor(Display *xdpy, Window appRootWindow, Cursor c, BOOL set)
{
  NSApplication *theApp = [NSApplication sharedApplication];
  NSWindow *w = [theApp mainWindow];
  Window window;

  if (w)
    window = ([XGContext _windowWithTag: [w windowNumber]])->ident;
  else
    window = appRootWindow;
  if (set)
    XDefineCursor(xdpy, window, c);
  else
    XUndefineCursor(xdpy, window);
}

#define ALPHA_THRESHOLD 158

Pixmap
xgps_cursor_mask(Display *xdpy, Drawable draw, const char *data, 
		  int w, int h, int colors)
{
  int j, i;
  unsigned char	ialpha;
  Pixmap pix;
  int bitmapSize = ceil((double)(w)/8.0) * h;
  char *aData = calloc(1, bitmapSize);
  char *cData = aData;

  if (colors == 4)
    {
      int k;
      for (j = 0; j < h; j++)
	{
	  k = 0;
	  for (i = 0; i < w; i++)
	    {
	      data += 3;
	      ialpha = (unsigned short)((char)*data++);
	      if (ialpha > ALPHA_THRESHOLD)
		*cData |= (0x01 << k);
	      k++;
	      if (k > 7)
		{
		  cData++;
		  k = 0;
		}
	    }
	  cData++;
	}
    }
  else
    {
      for (j = 0; j < bitmapSize; j++)
	*cData++ = 0xff;
    }

  pix = XCreatePixmapFromBitmapData(xdpy, draw, (char *)aData, w, h, 
				    1L, 0L, 1);
  free(aData);
  return pix;
}

Pixmap
xgps_cursor_image(Display *xdpy, Drawable draw, const char *data, 
		  int w, int h, int colors, XColor *fg, XColor *bg)
{
  int j, i, min, max;
  Pixmap pix;
  int bitmapSize = ceil((double)(w)/8.0) * h;
  char *aData = calloc(1, bitmapSize);
  char *cData = aData;

  min = 256;
  max = 0;
  if (colors == 4 || colors == 3)
    {
      int k;
      for (j = 0; j < h; j++)
	{
	  k = 0;
	  for (i = 0; i < w; i++)
	    {
	      int color = ((0.3*data[0]) + (0.59*data[1]) + (0.11*data[2]));
	      if (color > 128)
		*cData |= (0x01 << k);
	      if (color < min)
		{
		  min = color;
		  bg->red = (int)data[0] * 256; 
		  bg->green = (int)data[1] * 256; 
		  bg->blue = (int)data[2] * 256;
		}
	      else if (color > max)
		{
		  max = color;
		  fg->red = (int)data[0] * 256; 
		  fg->green = (int)data[1] * 256; 
		  fg->blue = (int)data[2] * 256;
		}
	      data += 3;
	      if (colors == 4)
		data++;
	      k++;
	      if (k > 7)
		{
		  cData++;
		  k = 0;
		}
	    }
	  cData++;
	}
    }
  else
    {
      for (j = 0; j < bitmapSize; j++)
	{
	  if ((unsigned short)((char)*data++) > 128)
	    *cData |= (0x01 << i);
	  cData++;
	}
    }
  
  pix = XCreatePixmapFromBitmapData(xdpy, draw, (char *)aData, w, h, 
				    1L, 0L, 1);
  free(aData);
  return pix;
}

- (void) DPShidecursor
{
  Pixmap shape, mask;
  XColor black, white;

  if (cursor_hidden)
    return;

  if (xgps_blank_cursor == None)
    {
      shape = XCreatePixmapFromBitmapData(XDPY, XDRW,
					  xgps_blank_cursor_bits, 
					  16, 16, 1, 0, 1);
      mask = XCreatePixmapFromBitmapData(XDPY, XDRW,
					 xgps_blank_cursor_bits, 
					 16, 16, 1, 0, 1);
      black.red = black.green = black.blue = 0;
      black = [self xColorFromColor: black];
      white.red = white.green = white.blue = 65535;
      white = [self xColorFromColor: white];
      
      xgps_blank_cursor = XCreatePixmapCursor(XDPY, shape, mask, 
					      &white, &black,  0, 0);
      XFreePixmap(XDPY, shape);
      XFreePixmap(XDPY, mask);
    }
  xgps_set_cursor(XDPY, XDRW, xgps_blank_cursor, YES);
  cursor_hidden = YES;
}

- (void) DPSshowcursor
{
  if (cursor_hidden)
    xgps_set_cursor(XDPY, XDRW, None, NO);
  cursor_hidden = NO;
}

- (void) DPSstandardcursor: (int)style : (void **)cid
{
  xgps_cursor_id_t *cursor;
  cursor = NSZoneMalloc([self zone], sizeof(xgps_cursor_id_t));
  switch (style)
    {
    case GSArrowCursor:
      cursor->c = XCreateFontCursor(XDPY, XC_left_ptr);     
      break;
    case GSIBeamCursor:
      cursor->c = XCreateFontCursor(XDPY, XC_xterm);
      break;
    default:
      cursor->c = XCreateFontCursor(XDPY, XC_left_ptr);     
      break;
    }
  cursor->defined = NO;
  if (cid)
    *cid = (void *)cursor;
}

- (void) DPSimagecursor: (float) hotx : (float) hoty : (float) w :  (float) h : (int)colors : (const char *)image : (void **)cid
{
  xgps_cursor_id_t *cursor;
  Pixmap source, mask;
  unsigned int maxw, maxh;
  XColor fg, bg;

  /* FIXME: We might create a blank cursor here? */
  if (image == NULL || w == 0 || h == 0)
    {
      *cid = NULL;
      return;
    }

  /* FIXME: Handle this better or return an error? */
  XQueryBestCursor(XDPY, XDRW, w, h, &maxw, &maxh);
  if (w > maxw)
    w = maxw;
  if (h > maxh)
    h = maxh;

  cursor = NSZoneMalloc([self zone], sizeof(xgps_cursor_id_t));
  source = xgps_cursor_image(XDPY, XDRW, image, w, h, colors, &fg, &bg);
  mask = xgps_cursor_mask(XDPY, XDRW, image, w, h, colors);
  bg = [self xColorFromColor: bg];
  fg = [self xColorFromColor: fg];

  cursor->c = XCreatePixmapCursor(XDPY, source, mask, &fg, &bg, 
				  (int)hotx, (int)(h - hoty));
  cursor->defined = NO;
  XFreePixmap(XDPY, source);
  XFreePixmap(XDPY, mask);
  if (cid)
    *cid = (void *)cursor;
}

- (void) DPSsetcursorcolor: (float)fr : (float)fg : (float)fb : (float)br : (float)bg : (float)bb : (void *)cid
{
  XColor xf, xb;
  xgps_cursor_id_t *cursor;

  cursor = (xgps_cursor_id_t *)cid;
  if (cursor->defined == NO)
    xgps_set_cursor(XDPY, XDRW, cursor->c, YES);

  /* Special hack: Don't set the color when fr < 0 */
  if (fr < 0)
    return;

  xf.red   = 65535 * fr;
  xf.green = 65535 * fg;
  xf.blue  = 65535 * fb;
  xb.red   = 65535 * br;
  xb.green = 65535 * bg;
  xb.blue  = 65535 * bb;
  xf = [self xColorFromColor: xf];
  xb = [self xColorFromColor: xb];
  XRecolorCursor(XDPY, cursor->c, &xf, &xb);
}


@end

